"""parse yarn ressource manager log4j aggregated logs"""
from dataclasses import dataclass, field
from typing import List, Pattern, Iterable, Pattern
import re
import logging
import arrow
from itertools import groupby

FILEPATTERN = 'yarn-yarn-resourcemanager*.log*'


@dataclass
class Parser():
    pattern_line: Pattern = field(init=False)

    def __post_init__(self):
        self.pattern_line = re.compile(
            r'(?P<ts>(\d+-?){3}\s(\d+[:,]?){4})\s(?P<level>[A-Z]+)\s{2}(?P<javaclass>\w+\.\w+)\s\(.*\((?P<java_id>\d+)\)\)\s-\s(?P<payload>.*)')

    def parse(self, data: Iterable[str]) -> List[dict]:
        return [self.parse_line(line) for line in data]

    def parse_line(self, logline: str) -> dict:
        """try to parse line, fail silently by returning an empty dict"""
        try:
            return self.pattern_line.match(logline).groupdict()
        except Exception as e:
            logging.debug(e)
            return {}


@dataclass
class StateProcessor:
    pattern_state: Pattern = field(init=False)

    def __post_init__(self):
        self.pattern_state = re.compile(
            r'(?P<id_application>application_\d+_\d+)(.*)from\s(?P<from_state>[A-Z_]+)\sto\s(?P<to_state>[A-Z_]+)'
        )

    def parse(self, data: List[dict]) -> List[dict]:
        "parse 'state' line(transition) and discard other"
        def merge(line):
            return {**{'ts': line['ts']},
                    **(self.parse_payload(line['payload']))}
        valid = [line for line in data if self.is_state(line)]
        return [merge(elt) for elt in valid]

    def parse_payload(self, payload: str) -> dict:
        """try to parse, fail silently by returning an empty dict"""
        try:
            return self.pattern_state.match(payload).groupdict()
        except Exception as e:
            logging.debug(e)
            return {}

    @staticmethod
    def is_state(parsed: dict) -> bool:
        # code 779
        try:
            return parsed['java_id'] == '779'
        except KeyError:
            return False

    @staticmethod
    def make_name(from_state: str, to_state: str) -> str:
        return f'{from_state.lower()}_to_{to_state.lower()}'

    @staticmethod
    def delta(states: List[dict], from_state: str, to_state: str) -> dict:
        """reduce as diff of time"""
        def diff(first, last) -> int:
            if (not first) or (not last):
                return None
            else:
                return abs((arrow.get(last)-arrow.get(first)).seconds)

        def get_state(state):
            try:
                return list(filter(lambda x: x['to_state'] == state, states))[0]
            except IndexError:
                return ''
        # 'to_state' is the last state known.
        # It is the current state of the app at time T
        t1 = get_state(from_state)
        t2 = get_state(to_state)
        try:
            return {'id_application': states[0]['id_application'],
                    'accept_to_running_ts': t2['ts'],
                    StateProcessor.make_name(from_state, to_state): diff(t1['ts'], t2['ts'])}
        except (IndexError, TypeError):
            return {}

    @staticmethod
    def process_diff(parsed: List[dict],
                     from_state: str,
                     to_state: str) -> List[dict]:
        result = []
        def keyfunc(x): return x['id_application']
        sort_data = sorted(parsed, key=keyfunc)
        for key, group in groupby(sort_data, keyfunc):
            result.append(StateProcessor.delta(
                list(group), from_state, to_state))
        return result


@dataclass
class QueueProcessor:
    pattern: Pattern = field(init=False)

    def __post_init__(self):
        self.pattern = re.compile(
            r'(.*)(appId:\s)(?P<id_application>application[0-9_]+)\s(user:\s(?P<user>[a-zA-Z0-9]+)).*(leaf-queue:\s(?P<queue>[a-zA-Z0-9_]+))([^#])(\#user-pending-applications+:\s?(?P<nb_user_pending>\d+))([^#])(#user-active-applications:\s?(?P<nb_user_active>\d+))([^#])(#queue-pending-applications:\s?(?P<nb_queue_pending>\d+))([^#])(#queue-active-applications:\s?(?P<nb_queue_active>\d+))')

    def parse_payload(self, payload: str) -> dict:
        try:
            found = self.pattern.match(payload).groupdict()
        except (AttributeError, TypeError):
            return {}
        for k in found.keys():
            if 'nb_' in k:
                found[k] = int(found[k])
        return found

    def parse_line(self, logline: dict) -> dict:
        return {**{'queue_state_ts': logline['ts']},
                **(self.parse_payload(logline['payload']))}

    @staticmethod
    def is_queue(parsed: dict) -> bool:
        try:
            return (parsed['java_id'] == '744') and (parsed['javaclass'] == 'capacity.LeafQueue')
        except KeyError:
            return False


def kpi_accept_to_run_parsed(parsed: List[dict]) -> List[dict]:
    processor = StateProcessor()
    return processor.process_diff(processor.parse(parsed),
                                  from_state='ACCEPTED',
                                  to_state='RUNNING'
                                  )


def kpi_accept_to_run(lines: List[str]) -> List[dict]:
    return kpi_accept_to_run_parsed(Parser().parse(lines))


def kpi_nb_queue_parsed(parsed: List[dict]) -> List[dict]:
    return [QueueProcessor().parse_line(line) for line in parsed if QueueProcessor.is_queue(line)]


def kpi_nb_queue(lines: List[str]):
    return kpi_nb_queue_parsed(Parser().parse(lines))

# FIXME: see  test & README
# also branch filter queue for poc on notebook


def is_equal(a: dict, b: dict)->bool:
    """compare record parsed on 'id_application' key"""
    try:
        return a['id_application'] == b['id_application']
    except (KeyError,TypeError):
        return False


def process(lines: Iterable[str]) -> List[dict]:
    parsed = Parser().parse(lines)
    nb_queue = kpi_nb_queue_parsed(parsed)
    accept_to_run = kpi_accept_to_run_parsed(parsed)
    merged = []
    for queue in nb_queue:
        for to_run in accept_to_run:
            try:
                if is_equal(queue, to_run):
                    merged.append({**queue, **to_run})
            except KeyError:
                pass
    return merged

    # add state
    # add payload
    # be sure to have dt for both
    # dataframe join ?
    # return nb_queue



from typing import List
import humanize
from datetime import timedelta, datetime
import pandas as pd
import arrow

DATA_DOC = {
    'id': 'yarn application ID',
    'user': 'who submitted this job',
    'name': 'job reference, for hive usually an ID',
    'startedTime': 'timestamp in milliseconds',
    'finishedTime': 'timestamp in milliseconds',
    'queue': 'yarn queue name for this job',
    'state': 'yarn state (ACCEPTED,RUN...)',
    'finalStatus': 'state in which this job stopped (FINISHED when everything is fine)',
    'progress': 'relevant only for RUNNING state',
    'trackingUI': 'from where the data are fetched. Could be History if the job is finished or ApplicationMaster if ongoing',
    'diagnostics': 'information for SUCCEDED or FAILED job. Sometimes useful for FAIL',
    'applicationType': 'TEZ,SPARK...',
    'elapsedTime': 'how long it tooks to finish the job in milliseconds',
}


def human_elapsed(nb: int) -> str:
    """milliseconds to human readable"""
    _t = humanize.i18n.activate("fr")
    return humanize.precisedelta(timedelta(milliseconds=nb), minimum_unit='hours')


def shift_ts(ts: int, tz) -> str:
    """convert UTC timestamp (ms) to europe/Paris tz in a standard string"""
    return arrow.get(ts, tzinfo='UTC').to(tz).format()


def flatten(raw: dict) -> list:
    return [elt for elt in raw['apps']['app']]


def to_minutes(milliseconds: int) -> int:
    return round((milliseconds/1000)/60)


def get_long_run(payload: dict, threshold: int, queue='') -> List[dict]:
    """
    Retrieve all application currently running for more than `threshold` minutes

    Parameters
    ----------
    payload:
        json returned py Yarn TimeServer
    threshold:
        number of minutes

    """
    if not payload:
        return []
    apps = flatten(payload)
    filtered = [elt for elt in apps if elt['state'] ==
        'RUNNING' and queue in elt['queue'].lower()]
    if not filtered:
        return []
    df = pd.DataFrame(filtered)
    df['elapsed_minutes'] = df['elapsedTime'].apply(to_minutes)
    df['started_time'] = df.startedTime.apply(
        lambda ts: shift_ts(ts, 'Europe/Paris'))
    df_threshold = df[df.elapsed_minutes > threshold]
    df_threshold['threshold_minutes'] = threshold
    fields = ['id',
              'user',
              'started_time',
              'queue',
              'elapsed_minutes',
              'progress',
              'allocatedMB',
              'allocatedVCores',
              'queueUsagePercentage',
              'threshold_minutes']
    return df_threshold[fields].to_dict(orient='records')



from dataclasses import dataclass, field
from pathlib import Path
import json
from pydantic import BaseModel, ValidationError
from typing import List, Optional
import stringcase
import toolz
from toolz.dicttoolz import keymap


class Capacity(BaseModel):
    # note: values are always in megaBytes
    memory_mb: int
    vcores_nb: int


class Queue(BaseModel):
    """queues object for a Leaf Yarn queue"""

    queue_name: str
    # Configured queue capacity in percentage relative to its parent queue
    capacity: float
    # Configured maximum queue capacity in percentage
    # relative to its parent queue
    max_capacity: float
    # The maximum number of applications this queue can have
    max_applications: Optional[int]
    max_applications_per_user: Optional[int]
    """Minimum User Percentage and User Limit Factor are ways to control
     how resources get assigned to users within the queues they are utilizing.
     The Min User Percentage is a soft limit on the smallest amount of resources 
     a single user should get access to if they are requesting it
    """
    user_limit: Optional[int]
    """User Limit Factor is set as a multiple of the queues minimum capacity
     where a user limit factor of 1 means the user can consume the entire minimum 
     capacity of the queue.
     """
    user_limit_factor: Optional[int]
    queues: Optional[List]


@dataclass
class Config:
    """Process results from REST API call on
    - `resourcemanager/v1/cluster/scheduler`
    - `resourcemanager/v1/cluster/metrics`

    more info on
    https://hadoop.apache.org/docs/current/hadoop-yarn/hadoop-yarn-site/ResourceManagerRest.html#Cluster_Information_API
    https://hadoop.apache.org/docs/current/hadoop-yarn/hadoop-yarn-site/ResourceManagerRest.html#Cluster_Scheduler_API
    """

    scheduler: dict
    metrics: dict
    queues: List[Queue] = field(init=False)
    cluster: Capacity = field(init=False)

    def __post_init__(self):
        self.queues = Config._proc_scheduler(self.scheduler)
        self.cluster = Config._proc_metrics(self.metrics)

    @classmethod
    def from_file(cls, schedulerjson: Path, metricsjson: Path):
        scheduler = json.load(schedulerjson.open())
        metrics = json.load(metricsjson.open())

        return cls(scheduler, metrics)

    def get(self, queue_name: str) -> Capacity:
        """capacity that should be always available"""
        try:
            queue = [elt for elt in self.queues if elt.queue_name == queue_name][0]
        except IndexError:
            raise ValueError(f"queue `{queue_name}` not found")

        return Capacity(
            memory_mb=round(self.cluster.memory_mb * queue.capacity / 100, 1),
            vcores_nb=round(self.cluster.vcores_nb * queue.capacity / 100, 1),
        )

    def get_queue(self,queuename)->Queue:
        try:
            return [queue for queue in self.queues if queue.queue_name==queuename][0]
        except IndexError:
            raise ValueError(f"queue `{queuename}` not found")

    def get_queue_maxcapacity(self, queuname: str):
        """capacity that can be used if there is extra free ressources"""
        pass

    @staticmethod
    def _proc_metrics(payload: dict) -> Capacity:
        config = payload["clusterMetrics"]
        return Capacity(
            memory_mb=config["totalMB"], vcores_nb=config["totalVirtualCores"]
        )

    @staticmethod
    def _proc_scheduler(payload: dict) -> List[Queue]:
        queues = Config._get_queues(payload)
        return [Config._proc_queue(queue) for queue in queues]

    @staticmethod
    def to_snake(queue: dict) -> dict:
        return toolz.dicttoolz.keymap(stringcase.snakecase, queue)

    @staticmethod
    def _proc_queue(queue: dict) -> Queue:
        clean = Config.to_snake(queue)
        if "queues" in clean:
            clean["queues"] = clean["queues"]["queue"]
        return Queue(**clean)

    @staticmethod
    def _get_queues(payload: dict) -> list:
        """extract list of queues from a scheduler json dump"""
        return payload["scheduler"]["schedulerInfo"]["queues"]["queue"]

    @staticmethod
    def _get_subqueue(queue: Queue, quename: str) -> Queue:
        if queue.queues:
            subqueues = [Queue(**(Config.to_snake(elt))) for elt in queue.queues]
            return [elt for elt in subqueues if elt.queue_name == quename][0]
        else:
            raise ValueError(f"no subqueue in {queue}")

    @staticmethod
    def _get_queue_raw(payload: dict, queuename: str) -> dict:
        try:
            queues = Config._get_queues(payload)
            return [queue for queue in queues if queue["queueName"] == queuename][0]
        except IndexError:
            return {}

    @staticmethod
    def _calculate_capacity(user_limit_factor: float, capacity: int) -> float:
        return user_limit_factor * capacity
